"""
Contains the "outer" part of the DNS resolver.
"""

import logging
import threading
import time

from util import Dyn


_LOGGER = logging.getLogger(__name__)


class DNSProxy(threading.Thread):
    """
    Provides an interface to the dynamic module that actually does the DNS
    proxying.
    """
    def __init__(self, parent):
        threading.Thread.__init__(self)
        self.setDaemon(True)

        self.parent = parent

        self.bind_addr = parent.config.get("dnsproxy", "bind")
        self.bind_port = parent.config.getint("dnsproxy", "port")
        self.bufsize = parent.config.getint("dnsproxy", "buffer_size")

        self.socketLock = threading.Lock()
        self.handlerLock = threading.Lock()

        # state for the request handlers
        self.handlerIndex = 0
        self.handlers = {}

        self.should_run = True

        self.dyn = Dyn.load_best_dyn()

        # check if we should try to configure ourselves as the primary DNS server
        if parent.config.getboolean("dnsproxy", "configure_dns"):
            platform = self.dyn.safeimport("namehelp/platform")
            platform.configureNamehelp()

        # load the latest Dyn; init and start the handler.
        self.updateModule()

    def run(self):
        import dns, socket

        # set up the listening socket
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.socket.settimeout(0.1)
            self.socket.bind((self.bind_addr, self.bind_port))
        except Exception as e:
            _LOGGER.critical("Could not bind port %d: %s" % (self.bind_port,
                str(e)))
            time.sleep(1)
            return

        while self.should_run:
            try:
                data, addr = self.socket.recvfrom(self.bufsize)

                # try to parse the data into a DNS query
                msg = None
                try: msg = dns.message.from_wire(data)
                except: continue

                with self.handlerLock:
                    self.handlers[self.handlerIndex]( (addr, msg) ) # enqueue

            except socket.timeout as e:
                pass
            except Exception as e:
                _LOGGER.info("got exception reading from socket: %s %s" % (type(e), str(e)))

        _LOGGER.warn("dnsproxy thread exiting")

    def shutdown(self):
        self.should_run = False
        self.socket.close()

    def updateServerList(self):
        with self.handlerLock:
            for i in self.handlers:
                self.handlers[i].updateServerList()

    def removeHandler(self, handlerIndex):
        with self.handlerLock:
            del self.handlers[handlerIndex]

    def updateModule(self):
        with self.handlerLock:
            self.dyn = Dyn.load_best_dyn()
            module = self.dyn.safeimport("namehelp")
            handler = self.dyn.safeimport("namehelp/dns/dnshandler")
            if module is None:
                self.handlerLock.release()
                _LOGGER.critical("tried to load DNS handler module, but failed. aborting.")
                return

            # if not first run
            if self.handlers:
                # tell the old one to stop
                self.handlers[self.handlerIndex]() # send empty message, signal to quit when possible

            # init and start the new version
            self.handlerIndex += 1
            self.handlers[self.handlerIndex] = handler.DNSHandler(self, self.handlerIndex)
            self.handlers[self.handlerIndex].start()

