import platform

__version_info__ = (1, 0)
__version__ = '.'.join(map(str, __version_info__))

import os, sys, time
import logging, logging.handlers

import util, remote_config
from dnsproxy import DNSProxy


class namehelp:
    """
    Main class for namehelp application.

    Instantiates (statically) several helper objects for network communication,
    time synchronization, and cryptograpy.

    Runs an updater thread that periodically checks for and downloads new
    versions of the DNS proxy handler and UI.

    Runs the DNS proxy daemon in a thread.

    Main thread control is passed to the selected UI toolkit (if available and
    enabled), which enters a run loop.
    
    Finally, it waits for a shutdown signal before cleaning up.
    """

    version = __version__
    versionTup = __version_info__

    platform = platform.system()

    instance = None

    config = None
    config_file = None
    orig_args = None

    remote_config = None

    updateAvailable = False
    updateVersion = None

    net = None
    crypto = None
    time = None

    updater = None
    dnsproxy = None

    runtime = None

    start_time_utc = None

    def __init__(self, config, config_file, orig_args):
        logging.info("namehelp version %s starting" % namehelp.version)

        namehelp.instance = self

        namehelp.config = config
        namehelp.config_file = config_file
        namehelp.orig_args = orig_args

        namehelp.net = util.NetHelper(self, namehelp.config)
        namehelp.crypto = util.CryptoHelper(namehelp.config)
        namehelp.time = util.TimeHelper(self, namehelp.config)

        namehelp.remote_config = remote_config.RemoteConfig(namehelp) # depends on nethelper, timehelper

        namehelp.updater = util.Updater(self)

        dyn = util.Dyn.load_best_dyn()
        runtime = dyn.safeimport("namehelp/runtime")
        namehelp.runtime = runtime.Runtime(self)

        namehelp.start_time_utc = namehelp.time.getUtcSecondsNow()

        if not config.has_option("general", "uuid") or not self.validUuid():
            # generate a UUID
            import uuid as uuidlib
            uuid = str(uuidlib.uuid4())
            logging.info("generated UUID: %s" % uuid)
            namehelp.config.set("general", "uuid", uuid)
            self.writeConfigFile()

        namehelp.dnsproxy = DNSProxy(self)

    def validUuid(self):
        import uuid as uuidlib
        try:
            uuidlib.UUID(namehelp.config.get("general", "uuid"))
        except ValueError:
            return False
        return True

    def writeConfigFile(self):
        namehelp.config.write(open(namehelp.config_file, "w"))

    def run(self):
        namehelp.dnsproxy.start()
        namehelp.updater.start()
        namehelp.runtime.start()

        while True:
            time.sleep(10)

    @staticmethod
    def setUpdateAvailable(version=None):
        flag = version is not None
        namehelp.updateAvailable = flag
        if version is None:
            namehelp.updateVersion = None
        else:
            namehelp.updateVersion = ".".join([str(x) for x in version])

        if namehelp.runtime:
            namehelp.runtime.setUpdateAvailable(version)

    def shutdown(self):
        namehelp.runtime.shutdown()
        namehelp.dnsproxy.shutdown()

    def restart(self):
        logging.info("shutting down")
        self.shutdown()

        logging.info("short sleep...")
        time.sleep(0.5)

        logging.info("calling os.execl")
        logging.shutdown()

        python = sys.executable
        os.execl(python, python, *namehelp.orig_args)
