"""
This module handles the interface with the AquaLab configuration servers.

All communication is done over HTTP(S)
"""

import datetime, urllib, json, logging

class RemoteConfig:

    def __init__(self, namehelp):
        """
        servers: list of base URLs to query
        """

        self.config = namehelp.config
        self.net = namehelp.net
        self.time = namehelp.time

        self.servers = self.config.get("remote_config", "servers").split(";")
        self.cache_time = self.config.getint("remote_config", "cache_seconds")
        self.prefix = self.config.get("remote_config", "query_prefix").strip("/")

        self.cache = {}
        self.expires = {}
        self.errors = {}
        self.server_used = {}
        self.headers = {}

    def get(self, key_tuple, cached=True):
        """
        Query the server for a configuration parameter.

        If cached is True (default), serve from the cache when possible.
        Setting cached to False forces a request
        """

        now = self.time.getUtcNow()

        have_valid = key_tuple in self.expires and now <= self.expires[key_tuple]

        if (not cached) or (not have_valid):
            # clean the data structures
            if key_tuple in self.cache: del self.cache[key_tuple]
            if key_tuple in self.headers: del self.headers[key_tuple]
            if key_tuple in self.expires: del self.expires[key_tuple]
            if key_tuple in self.errors: del self.errors[key_tuple]
            if key_tuple in self.server_used: del self.server_used[key_tuple]

            # do the request...try servers in order
            response = None
            server_used = None
            headers = {}
            errors = {}

            for server in self.servers:
                # make the URL
                url = "/".join([server.strip("/"), 
                                 self.prefix] + 
                                [urllib.quote(str(x)) for x in key_tuple])
                logging.debug("querying %s for %s: %s" % (server, str(key_tuple), url))

                try:
                    resp = None

                    # handle HTTP and HTTPS differently
                    if url.startswith("https://"):
                        # HTTPS
                        resp = self.net.doVerifiedHttps(url)
                        
                    elif url.startswith("http://"):
                        # HTTP
                        resp = self.net.doHttp(url)
                    else:
                        # unknown URL scheme. Skip it.
                        errors[server] = "unknown URL scheme"
                        continue

                    if resp is None:
                        errors[server] = "response is None"
                        continue

                    headers[server] = resp.info().headers

                    this_resp = resp.read()

                    # try to parse as JSON format. Fall back to string
                    try:
                        this_resp = json.loads(this_resp)
                    except Exception as e:
                        # silently fail to parse as JSON
                        pass

                    server_used = server
                    response = this_resp
                    break

                except Exception as e:
                    errors[server] = "%s: %s" % (type(e), str(e))

            # TODO use cache information from the headers, if we got a successful response.

            # update the cache structures
            self.expires[key_tuple] = now + datetime.timedelta(0, self.cache_time)
            self.errors[key_tuple] = errors

            # if we have a valid response, we store this info.
            if response is not None:
                self.cache[key_tuple] = response
                self.server_used[key_tuple] = server_used
                self.headers[key_tuple] = headers

        to_return = dict(query=key_tuple,
                         expires=self.expires[key_tuple],
                         errors=self.errors[key_tuple])

        # if we have a valid response
        if key_tuple in self.cache:
            to_return["server"] = self.server_used[key_tuple]
            to_return["headers"] = self.headers[key_tuple]
            to_return["response"] = self.cache[key_tuple]
            to_return["success"] = True
        else:
            to_return["success"] = False
            
        return to_return
