import httplib, socket, ssl, urllib2, rsa, urlparse, json, random, logging, threading
import dns.rdatatype, dns.query, dns.message, dns.exception, dns.flags
import datetime, calendar, time
import base64
import re, os
from ntplib import ntplib
from pydhcplib import dhcp_packet, dhcp_network
from traceback import format_exc

from portability import winifyPath


_LOGGER = logging.getLogger(__name__)
_MODMAP = dict()
_MODMAPLOCK = threading.RLock()

def clear_module_cache():
    with _MODMAPLOCK:
        _MODMAP.clear()

class ModuleLoadException(Exception):
    '''Exception when a namehelp module fails to load'''
    pass


class Dyn(object):
    """
    Verifies and loads dynamically-updated namehelp modules and resources.
    """

    fileMatch = re.compile(r"(?P<name>[^_]+)_(?P<channel>[^_]+)_(?P<version>[0-9\.]+)\.zip(?P<sig>\.signature)?")

    @staticmethod
    def versions():
        from main import namehelp
        local_dir = namehelp.config.get("updater", "versions_dir")
        if not os.path.exists(local_dir): return set()

        channel = namehelp.config.get("updater", "channel")

        name = namehelp.config.get("updater", "version_prefix")

        codeVersions = {}
        sigVersions = {}

        for f in os.listdir(local_dir):
            m = Dyn.fileMatch.match(f)
            if not m: continue

            if m.group("name") != name: continue
            if m.group("channel") != channel: continue

            version = parseVersion(m.group("version"))

            if m.group("sig"):
                sigVersions[version] = f
            else:
                codeVersions[version] = f

        have = set(codeVersions) & set(sigVersions)
        
        _LOGGER.debug("Dyn.versions have: %s" % str(have))

        return have

    @staticmethod
    def max_version():
        vers = Dyn.versions()
        if not vers: return None
        return max(vers)

    @staticmethod
    def load_best_dyn():
        """
        Try to verify and load available versions, in descending order,
        until we find one that succeeds.
        """

        versions = list(Dyn.versions())
        versions.sort(reverse=True)

        for version in versions:
            sv = strVersion(version)
            try:
                d = Dyn(version)
                d.load()
                _LOGGER.info('Loading Dyn version %s' % sv)
                return d
            except ModuleLoadException as e:
                _LOGGER.error('Version %s failed while loading: %s' %
                              (sv, str(e)))

        raise ModuleLoadException('All available versions failed to load')

    @staticmethod
    def load_best():
        """ DEPRECATED

        This function is unsafe. The importer returned by it may import modules
        multiple times, resulting in weird and unpleasant shit. Use
        load_best_dyn().safeimport for far better results.
        """
        return Dyn.load_best_dyn().importer

    def code_file(self):
        from main import namehelp
        local_dir = namehelp.config.get("updater", "versions_dir")
        name = namehelp.config.get("updater", "version_prefix")
        channel = namehelp.config.get("updater", "channel")
        version = strVersion(self.version)
        return os.sep.join([local_dir, "%s_%s_%s.zip" % (name, channel, version)])

    def sig_file(self):
        return "%s.signature" % self.code_file()

    def __init__(self, version):
        self.version = version
        self.importer = None

    def verify(self):
        from main import namehelp
        sig = self.sig_file()
        code = self.code_file()
        if not os.path.exists(sig):
            ModuleLoadException('Signature path "%s" does not exist' % sig)
        if not os.path.exists(code):
            ModuleLoadException('Code path "%s" does not exist' % code)

        sig_f = open(sig, 'rb')
        code_f = open(code, 'rb')
        sig_data = base64.b64decode(sig_f.read())
        sig_f.close()
        valid = namehelp.crypto.check_signature(code_f, sig_data)
        code_f.close()
        return valid

    def load(self):
        if not self.verify():
            msg = 'Namehelp module failed signature check %s' % self
            raise ModuleLoadException(msg)

        try:
            import zipimport
            importer = zipimport.zipimporter(self.code_file())

            module = importer.load_module("namehelp")
            if self.version != module.__version_info__:
                _LOGGER.warning("version mismatch: filename %s, module %s" %
                                (strVersion(self.version),
                                 strVersion(module.__version_info__)))

            from main import namehelp
            channel = namehelp.config.get("updater", "channel")

            if channel != module.__channel__:
                _LOGGER.error("channel mismatch: filename %s, module %s" % (channel, module.__channel__))
                raise RuntimeError("Dyn of wrong channel %s" % module.__channel__)

            self.importer = importer
        except Exception as e:
            raise ModuleLoadException('Exception while trying to load Dyn '
                                      'version %s from %s: %s %s' %
                                      (strVersion(self.version), self.code_file(), str(type(e)),
                                       str(e)))


    def safeimport(self, name):
        """ Load a dynamic module, ensuring each module is only loaded once.
        Platform-safe.

        Args:
            name: A string giving the name of a dynamic module, in/this/format.

        Returns:
            The requested module, which will have been freshly imported if it
            hadn't been imported previously.
        """
        module = None
        with _MODMAPLOCK:
            if name in _MODMAP:
                _LOGGER.info("Returning already loaded dynamic module %s" % name)
                module = _MODMAP[name]
            else:
                _LOGGER.info("Loading dynamic module %s" % name)
                try:
                    module = self.importer.load_module(winifyPath(name))
                except:
                    _LOGGER.critical("Exception while importing %s: %s" % (name, format_exc()))
                    raise ModuleLoadException(name)
                _MODMAP[name] = module
        return module


class Updater(threading.Thread):

    def __init__(self, root):
        threading.Thread.__init__(self)
        self.setDaemon(True)

        self.root = root

        self.local_dir = root.config.get("updater", "versions_dir")
        try: os.makedirs(self.local_dir)
        except: pass

        self.interval = root.config.getint("updater", "interval")
        self.randomness = root.config.getfloat("updater", "randomness")

        self.channel = root.config.get("updater", "channel")

        self.updaterLock = threading.Lock()

        self.getLatestVersion()

    def getLatestVersion(self):
        """
        Check for a newer version on the server.
        """

        _LOGGER.info("checking for latest version...")

        try:
            resp = self.root.remote_config.get( ("namehelp", "client_version", self.channel) )
            if not resp["success"]:
                raise RuntimeError("remote_config.get failed: %s" % str(resp["errors"]))

            versionInfo = resp["response"]
            if not versionInfo:
                raise RuntimeError("no version information")

            nhVersion = tuple(versionInfo)

            _LOGGER.info("latest client version is %s" % str(nhVersion))

            import main

            if nhVersion is not None and nhVersion > main.__version_info__:
                self.root.setUpdateAvailable(nhVersion)
                _LOGGER.info("updated client is available: %s" % str(nhVersion))
            else:
                self.root.setUpdateAvailable()

        except Exception as e:
            _LOGGER.error("exception while trying to get latest client version information: (%s) %s" % (type(e), str(e)))

        _LOGGER.info("checking for latest version...")

        # these are the versions we already have
        bestAvailable = Dyn.max_version()

        # now, find out which is the latest that is available
        try:
            resp = self.root.remote_config.get( ("namehelp", "dyn_version", self.channel) )
            if not resp["success"]:
                raise RuntimeError("remote_config.get failed: %s" % str(resp["errors"]))

            versionInfo = resp["response"]
            if not versionInfo:
                raise RuntimeError("no version information")

            version, codeFile, sigFile = max(versionInfo)
            version = tuple(version)

            base_url = resp["server"].strip("/")

            # handle non-absolute paths to the code or sig files
            if (not codeFile.startswith("http://")) and (not codeFile.startswith("https://")):
                codeFile = "/".join([base_url, codeFile.strip("/")])

            if (not sigFile.startswith("http://")) and (not sigFile.startswith("https://")):
                sigFile = "/".join([base_url, sigFile.strip("/")])

            _LOGGER.info("latest version is %s" % str(version))

            _LOGGER.info("code at %s" % codeFile)
            _LOGGER.info("sig at  %s" % sigFile)

            if version > bestAvailable:
                _LOGGER.info("latest version is newer, fetching")
                codeFilename = codeFile.rsplit("/", 1)[1]
                sigFilename = sigFile.rsplit("/", 1)[1]
                self.root.net.fetchFile(codeFile, os.sep.join([self.local_dir, codeFilename]))
                self.root.net.fetchFile(sigFile, os.sep.join([self.local_dir, sigFilename]))
                _LOGGER.info("got the newer version files")

        except Exception as e:
            _LOGGER.error("exception while trying to get latest version information: (%s) %s" % (type(e), str(e)))

        available = Dyn.versions()

        if not available: return None
        return max(available)

    def checkForUpdate(self):
        with self.updaterLock:
            current = Dyn.max_version()
            latest = self.getLatestVersion()

        return latest != current

    def run(self):
        from wranglib.cattle import report
        import main

        # startup reporting
        import platform
        snapshot = {}
        try:
            snapshot["sys"] = platform.system()
            snapshot["release"] = platform.release()
            snapshot["arch"] = platform.architecture()
            snapshot["mach"] = platform.machine()
            snapshot["proc"] = platform.processor()
            snapshot["python"] = platform.python_version_tuple()
            if platform.system() == "Windows":
                snapshot["Windows"] = platform.win32_ver()
            elif platform.system() == "Linux":
                snapshot["Linux"] = platform.linux_distribution()
            elif platform.system() == "Darwin":
                snapshot["Darwin"] = platform.mac_ver()
        except Exception:
            pass
        if snapshot:
            report("snapshot", snapshot)

        while True:
            # send heartbeat
            now = main.namehelp.time.getUtcSecondsNow()
            uptime = now - main.namehelp.start_time_utc
            report( "heartbeat", {"uptime":uptime, "now":now})

            # check for updates
            newVersion = self.checkForUpdate()
            _LOGGER.info("new version? %s" % newVersion)
            if newVersion:
                self.root.restart()

            # sleep for a while
            sleepTime = self.interval + (self.randomness * self.interval * 2 * (random.random() - 0.5))

            time.sleep(sleepTime)

def ipParse(ipStr):
    ip = []
    ipBytes = ipStr.strip().split(".")
    if len(ipBytes) != 4:
        raise ValueError("invalid ip -- <4 numbers")

    try:
        ip = [int(x) for x in ipBytes]
    except:
        raise ValueError("invalid ip -- not a number")

    for num in ip:
        if num < 0 or num > 255:
            raise ValueError("invalid ip -- number out of range")

    return tuple(ip)

def parseVersion(version):
    toks = version.split(".")
    newtoks = []
    for tok in toks:
        try: newtoks.append(int(tok))
        except: newtoks.append(tok)
    return tuple(newtoks)

def strVersion(version):
    return ".".join([str(x) for x in version])


# To minimize pain, deep parts of the code can access the singleton NetHelper
# through this.
_NETHELPER = None
def getNetHelper():
    assert _NETHELPER is not None, "NetHelper not instantiated!"
    return _NETHELPER


class NetHelper:
    """
    Functionality for:
        making DNS requests without relying on the OS's configured servers
        validating server certificate in HTTPS communication
            (when communicating with namehelp servers)
        fetching the list of namehelp reporting servers
        (re)choosing reporting servers
        monitoring for changes in the client's IP address *** TODO
            (i.e. used for triggering check of DHCP DNS servers)
    """

    class IPChangeMonitor(threading.Thread):
        def __init__(self, parent):
            threading.Thread.__init__(self)
            self.setDaemon(True)

            self.parent = parent

            self.start()

        def run(self):
            while True:
                self.parent._checkChangedIp()
                time.sleep(60)

    def __init__(self, parent, config):
        global _NETHELPER
        self.parent = parent
        self.config = config

        class VerifiedHTTPSConnection(httplib.HTTPSConnection):
            """
            Thanks: http://thejosephturner.com/blog/2011/03/19/
                https-certificate-verification-in-python-with-urllib2/
            """

            def connect(self):
                # overrides the version in httplib so that we do
                #    certificate verification
                sock = socket.create_connection((self.host, self.port),
                                                self.timeout)
                if self._tunnel_host:
                    self.sock = sock
                    self._tunnel()
                # wrap the socket using verification with the root
                #    certs in trusted_root_certs
                ca_file = config.get("general", "cert_file")

                self.sock = ssl.wrap_socket(sock,
                                            cert_reqs=ssl.CERT_REQUIRED,
                                            ca_certs=ca_file)

        class VerifiedHTTPSHandler(urllib2.HTTPSHandler):
            """
            Thanks: http://thejosephturner.com/blog/2011/03/19/
                https-certificate-verification-in-python-with-urllib2/
            """

            def __init__(self, connection_class=VerifiedHTTPSConnection):
                self.specialized_conn_class = connection_class
                urllib2.HTTPSHandler.__init__(self)
            def https_open(self, req):
                return self.do_open(self.specialized_conn_class, req)

        self.https_handler = VerifiedHTTPSHandler
        self.https_opener = urllib2.build_opener(self.https_handler)

        self.local_ip = self._getLocalIp()
        self.dhcp_servers = self.getDhcpDns()
        self.local_info_lock = threading.Lock()

        self.change_ip_thread = NetHelper.IPChangeMonitor(self)

        assert _NETHELPER is None, "NetHelper already instantiated!"
        _NETHELPER = self


    def _getLocalIp(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8",80))
        ip = s.getsockname()[0]
        s.close()
        return ip

    def _checkChangedIp(self):
        with self.local_info_lock:
            ip = self._getLocalIp()
            if ip != self.local_ip:
                # do a few things
                self.local_ip = ip
                self.dhcp_servers = self.getDhcpDns()
                if self.parent.dnsproxy is not None:
                    self.parent.dnsproxy.updateServerList()

    def _getRequest(self, url, data=None, headers={}):
        """
        Do DNS query to get the IP address to connect to,
        and create a urllib2.Request object with the
        Host header set properly.
        Also, set Content-Length if we are POSTing data.
        """

        urlp = urlparse.urlparse(url)

        toks = urlp.netloc.split(":")
        host = toks[0]

        # translate the hostname to an IP address with DNS
        # do this manually, since we can't rely on OS having DNS
        # configured properly
        resp = self._dnsQuery(host)

        addresses = self._dnsGetAddresses(resp)

        if not addresses: raise RuntimeError("no IP addresses for %s" % host)

        toks[0] = addresses[0]
        urlp = list(urlp)
        urlp[1] = ":".join(toks)
        urlp = tuple(urlp)

        url = urlparse.urlunparse(urlp)

        headers["Host"] = host

        if data is not None:
            headers["Content-Length"] = len(data)

        return urllib2.Request(url, data=data, headers=headers)

    def doHttp(self, url, data=None, headers={}):
        req = self._getRequest(url, data=data, headers=headers)
        timeout = self.config.getfloat("general", "http_timeout")
        return urllib2.urlopen(req, timeout=timeout)

    def doVerifiedHttps(self, url, data=None, headers={}):
        """
        Generic function to do HTTPS query and verify the certificate.
        """

        req = self._getRequest(url, data=data, headers=headers)
        timeout = self.config.getfloat("general", "http_timeout")
        return self.https_opener.open(req, timeout=timeout)

    def fetchFile(self, url, dest, buf_size=65536):
        f = open(dest, "wb")
        try:
            rf = self.doHttp(url)

            while True:
                data = rf.read(buf_size)
                if data:
                    f.write(data)
                else:
                    break
        except Exception as e:
            raise e
        finally:
            f.close()

    def simpleDnsLookup(self, query):
        resp = self._dnsQuery(query)
        addresses = self._dnsGetAddresses(resp)
        if addresses: return addresses[0]
        return None

    def _dnsQuery(self, query, rdtype=dns.rdatatype.A):
        default_dns = self.config.get("general", "default_dns")
        timeouts = [float(x) for x in self.config.get("general", "dns_timeouts").split(",")]

        for timeout in timeouts:
            try:
                message = dns.message.make_query(query, rdtype)
                resp = dns.query.udp(message, default_dns, timeout=timeout)
                if "TC" in dns.flags.to_text(resp.flags):
                    message = dns.message.make_query(query, rdtype)
                    resp = dns.query.tcp(message, default_dns, timeout=timeout)
                return resp
            except dns.exception.Timeout as e:
                _LOGGER.debug("got DNS timeout exception: %s" % str(e))
            except Exception as e:
                _LOGGER.error("got exception while doing DNS lookup: %s" % str(e))

        raise RuntimeError("DNS lookup (rdtype=%s) failed for %s" % (dns.rdatatype.to_text(rdtype), query))

    def _dnsGetAddresses(self, message):
        addresses = []
        for rrset in message.answer:
            if rrset.rdtype != dns.rdatatype.A: continue
            for rr in rrset:
                addresses.append(str(rr.address))
        return addresses

    def getDhcpDns(self, timeout=2, non_standard_attempts=1):

        # checked the spec for DHCP, responses to DHCPINFORM messages
        # shouldn't have to come from the default client port. Still,
        # let's try that one first, and if it fails to bind, then we'll
        # try additional random port.

        listen_address = "0.0.0.0"
        server_port = 67
        client_port = 68

        for port in [client_port] + [random.randint(1025, 65535) for _ in xrange(non_standard_attempts)]:

            _LOGGER.debug("trying to bind dhcp client to port %d" % port)

            client = dhcp_network.DhcpClient(listen_address, port, server_port)
            client.BindToAddress()

            if client.Errors():

                _LOGGER.debug("failed to bind dhcp client to port %d; errors: %s" % (port, str(client.Errors())))

                client.Close()

                continue

            _LOGGER.debug("bound dhcp client to local port %d" % port)

            # if we get here, we have a valid DHCP client instance.
            # Let's try to send some stuff.

            ip = self._getLocalIp()

            ipTup = [int(x) for x in ip.split(".")]
            #print ip, ipTup

            xidToMask = {}

            # since we don't know the broadcast address, send to a bunch of broadcast addresses.
            for mask in xrange(8, 25):
                xid = tuple([random.randint(0,255), random.randint(0,255), random.randint(0,255), random.randint(0,255)])
                xidToMask[xid] = mask

                p = dhcp_packet.DhcpPacket()
                p.SetOption("dhcp_message_type", [8])
                p.SetOption("op", [1])
                p.SetOption("hops", [0])
                p.SetOption("xid", xid)
                p.SetOption("ciaddr", ipTup)
                p.SetOption("parameter_request_list", [6])
                p.SetOption("htype", [1])
                p.SetOption("hlen", [6])

                # compute the destination IP
                ipInt = sum([2**(index*8) * x for index, x in enumerate(list(reversed(ipTup)))])
                broadcast = ipInt | (2**mask - 1)
                current = broadcast
                newTup = []
                for i in xrange(4):
                    newTup.insert(0, current % 256)
                    current /= 256
                broadcastIp = ".".join([str(x) for x in newTup])

                #print ipTup, ipInt, broadcast, newTup, broadcastIp

                try:
                    client.SendDhcpPacketTo(p, broadcastIp, server_port)
                except Exception as e:
                    _LOGGER.info("problem sending DHCP probe to %s: (%s) %s" % (broadcastIp, str(type(e)), str(e)))

            #expires = time.time() + timeout
            #while True:
            #    now = time.time()
            #    if now >= expires: break
            #
            #    p2 = client.GetNextDhcpPacket(expires - now)
            #
            #    print p2

            p2 = client.GetNextDhcpPacket(timeout)

            # clean up...
            client.Close()

            if p2 is not None:
                #rxid = p2.GetOption("xid")
                #print rxid, xidToMask.get(tuple(rxid), None)
                #serv = p2.GetOption("server_identifier")
                #print "DHCP server", serv
                #gw = p2.GetOption("giaddr")
                #print "Gateway", gw
                dnslist = p2.GetOption("domain_name_server")

                dnses = []
                while len(dnslist) >= 4:
                    dnses.append(".".join([str(x) for x in dnslist[:4]]))
                    del dnslist[:4]

                _LOGGER.debug("got DHCP DNS server list: %s" % str(dnses))

                return dnses

        _LOGGER.debug("failed to get DHCP DNS server list")

        return None

class CryptoHelper:
    """
    Implement some cryptographic helper functionality:
        verify signature of data/contents of file
    """

    def __init__(self, config):
        self.config = config

        pub_key_data = open(config.get("general", "public_key_file"),
                'rb').read()
        self.pub_key = rsa.PublicKey.load_pkcs1(pub_key_data)

    def check_signature(self, data_or_file, sig):
        """
        Return whether signature 'sig' is valid for the given
        data or data in a file, using the public RSA key.
        """
        try:
            rsa.verify(data_or_file, sig, self.pub_key)
        except rsa.VerificationError:
            _LOGGER.warning('RSA verify failed')
            return False
        return True

class TimeHelper(threading.Thread):

    def __init__(self, parent, config):
        threading.Thread.__init__(self)
        self.setDaemon(True)

        self.parent = parent

        self.server = config.get("general", "default_ntp_server")
        self.interval = config.getint("general", "ntp_interval")

        self.lock = threading.Lock()

        self.c = ntplib.NTPClient()
        self.state = None
        self.offset = 0.0

        self.start()

    def run(self):
        while True:
            try:
                ip = self.parent.net.simpleDnsLookup(self.server)
                newState = self.c.request(ip, version=3)
                with self.lock:
                    self.state = newState
                    self.offset = self.state.offset
                    _LOGGER.info("NTP success: offset %0.3f" % self.offset)
            except Exception as e:
                _LOGGER.error("NTP error: %s" % str(e))
            time.sleep(self.interval)

    def getOffset(self):
        return self.offset

    def printState(self):
        with self.lock:
            print('Version number : %d' % self.state.version)
            print('Offset : %f' % self.state.offset)
            print('Stratum : %s (%d)' % (ntplib.stratum_to_text(self.state.stratum), self.state.stratum))
            print('Precision : %d' % self.state.precision)
            print('Root delay : %f ' % self.state.root_delay)
            print('Root dispersion : %f' % self.state.root_dispersion)
            print('Delay : %f' % self.state.delay)
            print('Leap indicator : %s (%d)' % (ntplib.leap_to_text(self.state.leap), self.state.leap))
            print('Poll : %d' % self.state.poll)
            print('Mode : %s (%d)' % (ntplib.mode_to_text(self.state.mode), self.state.mode))
            print('Transmit timestamp : ' + time.ctime(self.state.tx_time))
            print('Reference timestamp : ' + time.ctime(self.state.ref_time))
            print('Original timestamp : ' + time.ctime(self.state.orig_time))
            print('Receive timestamp : ' + time.ctime(self.state.recv_time))
            print('Destination timestamp : ' + time.ctime(self.state.dest_time))
            print('Reference clock identifier : ' + ntplib.ref_id_to_text(self.state.ref_id, self.state.stratum))

    def getUtcNow(self):
        return datetime.datetime.utcnow() + datetime.timedelta(0, self.offset)

    def getUtcSecondsNow(self):
        dt = datetime.datetime.utcnow()
        # calendar.timegm loses sub-second resolution
        return calendar.timegm(dt.utctimetuple()) + (float(dt.microsecond)/1000000) + self.offset


"""
Scratch directory management. We need a private, persistent directory in which
to store state and other cross-launch data.
"""
_SCRATCH_DIR = './scratch'


def getscratchdir(subdir=""):
    """ Return the currently configured scratch directory. This location is
    writable and persistent across runs of the program. """
    # TODO: Make sure this dir is usable.
    sdir = os.path.join(_SCRATCH_DIR, subdir)
    if not os.path.exists(sdir):
        try:
            os.makedirs(sdir)
        except OSError:
            pass
    return sdir

