#!/usr/bin/env python2.7
"""
namehelp, an AquaLab Project
"""

# Required for dyn.namehelp.experimenter.
from __future__ import division

__author__ = 'jotto@eecs.northwestern.edu (John Otto)'

import json
import os
import sys
import traceback
import logging, logging.handlers
from signal import signal, SIGINT

if "darwin" in sys.platform:
    sys.path.insert(0, "./lib/python2.7")

from libnamehelp.main import namehelp
from libnamehelp.util import getscratchdir
from libnamehelp.portability import onWindows
from wranglib.cattle import init as initcattle, shutdown as shutdowncattle
from wranglib.cattle import report

_namehelp_instance = None

def main():
    # Check to see if we're running in a windows exe
    if hasattr(sys, 'frozen') and sys.frozen in ('windows_exe', 'console_exe'):
        install_path = os.path.abspath(sys.executable)
    else:
        install_path = os.path.realpath(__file__)
    if onWindows():
        path_sep = "\\"
    else:
        path_sep = "/"
    # Change to the install directory, making life easier for init scripts.
    install_dir = path_sep.join(install_path.split(path_sep)[:-1])
    os.chdir(install_dir)

    # save the original arguments, to use if we have to restart the program
    orig_args = list(sys.argv)

    # set up logging
    logger = logging.getLogger()
    loghandler = logging.handlers.TimedRotatingFileHandler("namehelp_log", when='d', interval=1, backupCount=7)
    loghandler.setFormatter(logging.Formatter(fmt='%(asctime)s\t%(levelname)s:%(name)s\t%(message)s', datefmt='%m/%d/%Y %I:%M:%S %p'))
    logger.addHandler(loghandler)
    logger.critical("namehelp starting with args: %s" % orig_args)

    # parse command-line arguments
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('--config', default='namehelp.cfg',
                      help='Read config from CONFIG.')
    parser.add_option('-d', action='store_true', dest='daemonize',
                      default=False, help='Daemonize.')
    parser.add_option('-p', action='store', dest='pidfile',
                      default=None, help='Path to the PID file.')
    opts, args = parser.parse_args()

    if opts.daemonize:
        bg_args = [arg for arg in orig_args if arg != '-d']
        background(bg_args)

    # If given one on the command line, open a pidfile. This should not be reached on Windows.
    pidfile = None
    if opts.pidfile != None:
        pidfile = open(opts.pidfile, 'wb')
        pidfile.write(str(os.getpid()))
        pidfile.flush()

    # load the configuration file
    import ConfigParser
    config = ConfigParser.ConfigParser()
    if not os.path.exists(opts.config):
        config.read("namehelp.default.cfg")
        config.write(open(opts.config, "w"))
    else:
        config.read(opts.config)
    logger.setLevel(getattr(logging, config.get("logging", "level").upper()))
    if not config.getboolean("logging", "quiet"):
        logger.addHandler(logging.StreamHandler())

    # Set up reporting.
    cattlescratch = getscratchdir('reports')
    initcattle(project="namehelp", workingdir=cattlescratch, port=8088)

    try:
        signal(SIGINT, lambda signum, frame: shutdown(pidfile))
    except ValueError:
        # Python signals can only be set in main thread. If we are not in the
        # main thread, don't set signal handlers
        logging.warn("Not in main thread, can't set signal handlers")

    # If this is only scoped within the try, exceptions don't get printed. The
    # hell?
    _namehelp_instance = None
    try:
        _namehelp_instance = namehelp(config, opts.config, orig_args)
        _namehelp_instance.run()
        logging.debug('Run finished')
    except (KeyboardInterrupt, SystemExit):
        logging.debug('Exit by KeyboardInterrupt or SystemExit')
    except Exception as e:
        if _namehelp_instance is not None:
            errreport = dict()
            errreport['type'] = str(type(e))
            errreport['traceback'] = traceback.format_exc()
            report('errors', json.dumps(errreport), async=False)
        logging.error(str(e))
        traceback.print_exc()
    finally:
        shutdown(pidfile)


def background(args):
    python = sys.executable
    os.system(python + " " + " ".join(args) + " >/dev/null 2>/dev/null&")
    sys.exit(0)


def shutdown(pidfile):
    """ Stop all components so we can exit cleanly.

    Args:
        pidfile: The pidfile to close, or None.
    """
    logging.warn('Shutting down')
    if _namehelp_instance is not None:
        logging.debug('Namehelp instance shutting down')
        _namehelp_instance.shutdown()
    shutdowncattle()
    logging.shutdown()
    if pidfile is not None:
        pidfile.close()
        try:
            os.unlink(pidfile.name)
        except OSError:
            pass
    sys.exit(0)


if __name__ == "__main__":
    # Prevents an uncaught UnicodeEncodeError deep in many versions of sgmllib
    # that causes BeautifulSoup to crash.
    # site.py removes setdefaultencoding from sys. reload() puts it back.
    reload(sys)
    sys.setdefaultencoding("utf-8")
    main()

# vim: set ts=4 sw=4 tw=0 :
