#!/usr/bin/python2.6
""" The client-side library for wrangler.

Import init and report and be sure to call the former before you try to use the
latter.
"""

import cPickle as pickle
import json
import logging
import re
import socket
from httplib import HTTPConnection, HTTPSConnection
from os import path
from Queue import Queue
from random import randint
from threading import Thread
from time import sleep, time
from uuid import uuid4 as uuid


_INIT = False
_DEBUG = False
_AUTOKEYS = {"+tstamp": time}
_LOGGER = logging.getLogger("cattle")
_CATTLE = None
_QUEUE = Queue()
_REPORTER = None
_PAT_RESOURCE = "^/?[\w\-]+$"
_PAT_PROJECT = _PAT_RESOURCE


class _Cattle(object):
    """ Module state object. """
    # With exponential backoff, the last attempt will be a little over five
    # days after the first.
    reportbaseinterval = 30
    reportmaxattempts = 13

    def __init__(self):
        self.client_uuid = str(uuid())
        self.host = "reports.aqualab.cs.northwestern.edu"
        self.port = 80
        self.use_ssl = True
        self.project = None
        self.workingdir = None
        self.reporting_enabled = True
        self.reports = []

    def __str__(self):
        if self.use_ssl:
            proto = "https"
        else:
            proto = "http"
        return "client %s pointed at %s://%s:%d/%s/* with %d saved reports" % \
                (self.client_uuid,
                 proto,
                 self.host,
                 self.port,
                 self.project,
                 len(self.reports))


class _Report(object):
    """ Wraps up an individual report for argument-passing and serialization.
    """
    def __init__(self):
        self.resource = None
        self.data = None
        self.attempts = 0
        self.last_attempt = 0

    def __str__(self):
        return "report pointed at %s carrying %s (attempts: %d, last: %d)" % \
                (self.resource, self.data, self.attempts, self.last_attempt)


class _Reporter(Thread):
    """ Periodically attempts to send all queued reports. Performs exponential
    backoff. """
    def __init__(self):
        Thread.__init__(self)
        self.daemon = True
        self.should_run = True
        self.interval = 30
        self.last_run = time() + randint(0, self.interval)

    def run(self):
        def nextrun(rep):
            base = float(randint(190, 210)) / 100
            multiplier = base ** rep.attempts
            next_attempt = rep.last_attempt
            next_attempt += _Cattle.reportbaseinterval * multiplier
            return next_attempt

        _dlog("Reporter thread started.")
        while self.should_run:
            # This thread is the only consumer of _QUEUE, so the worst that
            # will happen here is that a report won't be pulled out of the
            # queue for one extra iteration of this loop.
            process_immediately = False
            while not _QUEUE.empty():
                rep = _QUEUE.get()
                _CATTLE.reports.append(rep)
                _dlog("Pulled off the queue: %s" % rep)
                process_immediately = True

            if process_immediately or time() > self.last_run + self.interval:
                keepreports = []
                if _CATTLE.reports:
                    _dlog("Processing reports...")
                for r in _CATTLE.reports:
                    if not self.should_run:
                        break
                    if time() > nextrun(r):
                        _dlog("Due: %s" % r)
                        success, result = _sendreport(r)
                        # We want to keep any report that fails for a transient
                        # reason (i.e., not a wrangler protocol problem) and
                        # hasn't been tried the maximum number of times.
                        keep = False
                        if not success:
                            if 'response' in result and not result['response']:
                                msg = "Sending report %s: got response: %s" % \
                                        (r, result['response'])
                                _LOGGER.error(msg)
                            else:
                                if r.attempts < _Cattle.reportmaxattempts:
                                    keep = True
                        if success:
                            _dlog("Report successfully sent.")
                        else:
                            _LOGGER.info("Error sending report: %s" % result)
                            if not keep:
                                msg = "Abandoning report after %d tries." % \
                                        r.attempts
                                _LOGGER.warning(msg)
                        if keep:
                            keepreports.append(r)
                    else:
                        due_in = int(nextrun(r) - time() + 1)
                        _dlog("Due in %d s: %s" % (due_in, r))
                _CATTLE.reports = keepreports
                _savestate(_CATTLE.workingdir)
                self.last_run = time()
            sleep(1)
        _dlog("Reporter thread finished.")

    def shutdown(self):
        self.should_run = False


class ReportError(Exception):
    pass


class CattleInitError(Exception):
    pass


def _picklepath(workingdir):
    return path.join(workingdir, "cattle.pickle")


def _savestate(workingdir):
    with open(_picklepath(workingdir), 'wb') as pckl:
        pickle.dump(_CATTLE, pckl, protocol=-1)


def _loadstate(workingdir):
    global _CATTLE
    with open(_picklepath(workingdir), 'r') as pckl:
        _CATTLE = pickle.load(pckl)


def _dlog(msg):
    if _DEBUG:
        _LOGGER.debug(msg)


def _sendreport(rep):
    """ Inner, synchronous report method. Updates the bookkeeping in _Report
    instances.

    Args:
        rep: An instance of Report.

    Returns:
        If async == True, None.
        Else, a 2-tuple of a boolean indicating end-to-end success or failure
        and a dict containing:
            "response": The JSON object (loads'd to a dict) returned by
                wrangler.
            "http_code": The http code returned by wrangler.
            "exception": Any exception encountered during the report, other
                than during argument validation.
    """
    success = False
    result = {}
    if _CATTLE.use_ssl:
        http_con = HTTPSConnection(_CATTLE.host, _CATTLE.port)
    else:
        http_con = HTTPConnection(_CATTLE.host, _CATTLE.port)
    try:
        http_con.request("POST", rep.resource, rep.data)
    except socket.error as se:
        result["exception"] = se

    if "exception" not in result:
        response = http_con.getresponse()
        result["http_code"] = response.status
        if response.status == 200:
            wrangler_reply = response.read()
            try:
                result["response"] = json.loads(wrangler_reply)
            except ValueError:
                result["exception"] = ReportError(
                    "Couldn't parse reply from server: %s" % wrangler_reply)
        if "response" in result and not result["response"]:
            success = True

    rep.attempts += 1
    rep.last_attempt = time()
    return success, result


def report(resource, data, async=True, project=None, _dryrun=False):
    """ Submit data to the given http resource. If cattle has been globally
    disabled, this method does nothing and returns as if it had successfully
    delivered or queued the data.

    Args:
        resource: The absolute path of the resource to submit to. (e.g.,
            /dummy)
        data: A dict that will be passed to json.dumps(). Any object that JSON
            can't represent will go in as str(obj), so make sure that's what
            you want.
        async: If True, this call will put data into a queue for reliable
            delivery and return None immediately. If False, this call will
            block and return the results of the attempt. The latter is mostly
            useful for debugging.
        project: If not None, override the project passed to init.

    Returns:
        If async == True, None.
        Else, a 2-tuple of a boolean indicating end-to-end success or failure
        and a dict containing:
            "response": The JSON object (loads'd to a dict) returned by
                wrangler.
            "http_code": The http code returned by wrangler.
            "exception": Any exception encountered during the report, other
                than during argument validation.

    Raises:
        ReportError: If init() has not been called, the given resource is
        nonsense, or the data isn't acceptable as defined above.
    """
    if not _INIT:
        raise ReportError("init() not called")

    if not _CATTLE.reporting_enabled:
        _dlog("cattle disabled: dropping report at %s/%s: %s" % (project,
              resource, data))
        if async:
            return None
        else:
            return True, { "response": dict(),
                           "http_code": 200,
                           "exception": None }

    if not isinstance(resource, str):
        raise ValueError("resource must be a string")
    if not isinstance(data, dict):
        raise ValueError("data must be a dictionary")
    if project is None:
        project = _CATTLE.project
    if not re.match(_PAT_PROJECT, project):
        raise ValueError("project must match %s" % _PAT_PROJECT)
    if not re.match(_PAT_RESOURCE, resource):
        raise ValueError("resource must match %s" % _PAT_RESOURCE)

    resource = project + "/" + resource
    while resource.find("//") >= 0:
        resource = resource.replace("//", "/")
    while resource and resource[-1] == "/":
        resource = resource[0:len(resource) - 1]

    # Add automatic client-side keys.
    data['+client'] = _CATTLE.client_uuid
    for key, value in _AUTOKEYS.items():
        data[key] = value()
    if _DEBUG:
        data['+debug'] = True

    try:
        json_data = json.dumps(data, default=lambda x: str(x))
    except TypeError as te:
        raise ReportError("json.dumps raised: %s" % str(te))

    rep = _Report()
    rep.resource = resource
    rep.data = json_data

    # For testing purposes.
    if _dryrun:
        return rep

    if async:
        _QUEUE.put(rep)
        return None
    else:
        return _sendreport(rep)


def getclientid():
    """ Get the unique ID that this client reports as.

    Returns:
        The client's unique ID, represented as a string.
    Raises:
        CattleInitError: If init() has not been called.
    """
    if not _INIT:
        raise CattleInitError("init() was not called")
    return _CATTLE.client_uuid


def init(host=None, port=None, use_ssl=None, project=None, workingdir=None,
         autokeys=None, debug=False):
    """ Prepare the module for reporting and start the thread responsible for
    reliable delivery.

    Args:
        host: A string. Use to override the default.
        port: An integer. Use to override the default.
        use_ssl: A boolean. Defaults to True.
        project: The name of the project, as configured on the server side.
        autokeys: A dictionary mapping strings to callables. cattle will add
            each key to *every* report, invoking its value and putting that
            into the final report. Use this judiciously.
        debug: A boolean. If true, cattle will log copious information via the
            logging module at level DEBUG.
    """
    global _INIT, _DEBUG, _LOGGER, _CATTLE, _REPORTER

    if project is None:
        raise CattleInitError("project not specified")
    if not re.match("^\/?[\w\d\-\_\.]+$", project):
        raise CattleInitError("invalid value for project: %s" % project)
    if project[0] != "/":
        project = "/" + project
    if workingdir is None:
        raise CattleInitError("workingdir not specified")
    if not path.isdir(workingdir):
        raise CattleInitError("workingdir must be a directory")
    try:
        open(_picklepath(workingdir), 'ab').close()
    except IOError:
        raise CattleInitError("cannot write to workingdir")

    if debug:
        _DEBUG = True
    if autokeys is not None:
        for key, value in autokeys.items():
            if not hasattr(value, "__call__"):
                raise CattleInitError(
                        "value for %s in autokeys is not callable" % key)
            _AUTOKEYS[key] = value
    try:
        _loadstate(workingdir)
    except (IOError, EOFError):
        _dlog("Creating new state object.")
        _CATTLE = _Cattle()
    if host is not None:
        _CATTLE.host = host
    if port is not None:
        _CATTLE.port = port
    if use_ssl is not None:
        _CATTLE.use_ssl = use_ssl
    _CATTLE.project = project
    _CATTLE.workingdir = workingdir
    _savestate(workingdir)
    _dlog("Initializing with config: %s" % _CATTLE)

    if _REPORTER is None:
        _REPORTER = _Reporter()
        _REPORTER.start()
    _INIT = True


def shutdown(block=True):
    """ Stop the reporting thread, ensuring all data is either safely home or
    safely on disk.

    Args:
        block: A boolean. Whether this function should .join() the reporting
            thread before returning or return immediately after telling it to
            stop.
    """
    _REPORTER.shutdown()
    if block:
        _REPORTER.join()


def is_enabled():
    """ Return the value of the global on/off switch for the cattle library.
    """
    if not _INIT:
        raise ReportError("init() not called")
    return _CATTLE.reporting_enabled


def set_enabled(enabled):
    """ Set the value of the global on/off switch for the cattle library. If
    called with False, this function puts the module in a state where report()
    discards its arguments and indicates success in its return value. This
    makes it possible to disable all reporting without having to wrap every
    single call to report() in a branch. """
    if not _INIT:
        raise ReportError("init() not called")
    _dlog("Setting global on/off to %s." % str(enabled))
    _CATTLE.reporting_enabled = enabled

